"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
// @ts-ignore
const uuidProxy = require("simple-random-id");
const ee_1 = require("./ee");
const transition_1 = require("./transition");
const types_1 = require("./types");
var types_2 = require("./types");
exports.PipeFlags = types_2.PipeFlags;
exports.StateStructFields = types_2.StateStructFields;
exports.TransitionStepTypes = types_2.TransitionStepTypes;
exports.TransitionStepFields = types_2.TransitionStepFields;
exports.MutationTypes = types_2.MutationTypes;
exports.StateRelations = types_2.StateRelations;
var transition_2 = require("./transition");
exports.Transition = transition_2.default;
// TODO thank you rollup for years of linking issues
const uuid = uuidProxy.default || uuidProxy;
const assert = function (cond, msg) {
    if (!cond) {
        throw Error(msg);
    }
};
function machine(states = [], constructor) {
    var instance = new (constructor || AsyncMachine)();
    if (states instanceof Array) {
        for (let state of states) {
            instance[state] = {};
            instance.register(state);
        }
    }
    else {
        for (let state of Object.keys(states)) {
            instance[state] = states[state];
            instance.register(state);
        }
    }
    return instance;
}
exports.machine = machine;
/**
 * Base class to extend. Define states as prototype attributes or
 * inside of the constructor. In the latter case remember to call
 * this.registerAll() afterwards (for every sub constructor).
 *
 * The [[Exception]] state is already provided.
 *
 * ```
 * class FooStates extends AsyncMachine {
 *   Enabled: {}
 *
 *   Downloading: {
 *     drop: 'Downloaded'
 *   }
 *
 *   Downloaded = {
 *     drop: 'Downloading'
 *   }
 *
 *   constructor(target) {
 *     super(target)
 *     this.registerAll()
 *   }
 * }
 *
 * class Foo {
 *   constructor() {
 *     this.states = new FooStates(this)
 *   }
 *
 *   Downloading_state(states, url) {
 *   	 fetch(url, this.states.addByCallack('Downloaded'))
 *   }
 *
 *   Downloaded_state(states, local_path) {
 *   	 console.log(`Downloaded ${this.url} to ${local_path}`)
 *   }
 * }
 * ```
 */
class AsyncMachine extends ee_1.default {
    /**
     * Creates a new instance with only one registered state - Exception.
     *
     * When extending the class, you should register your states by using either
     * the [[registerAll]] or [[register]] methods at the end of every sub
     * constructor.
     *
     * @param target Target object for the transitions, useful when composing a
     *   machine.
     * @param register_all Automatically registers all defined states. Works
     *   only in case there no other attributes defined on the prototype.
     * @see [[AsyncMachine]] for the usage example.
     */
    constructor(target, register_all = true) {
        super();
        /**
         * Exception state's properties. See [[Exception_state]] final transition
         * handler.
         * @type {IState}
         */
        this.Exception = {
            multi: true
        };
        /** List of all registered state names */
        this.states_all = [];
        /**
         * Promise created by one of the delayed mutation methods:
         * - `...ByCallback`
         * - `...ByListener`
         *
         * Resolved once the callback/listener gets called.
         */
        this.last_promise = null;
        /** Map of piped states and their targets. */
        this.piped = {};
        /**
         * If true, an exception will be printed immediately after it's thrown.
         * Automatically turned on with logLevel > 0.
         */
        this.print_exception = false;
        /** List of active states. */
        this.states_active = [];
        this.queue_ = [];
        /** If true, this machine is currently during a transition. */
        this.lock = false;
        /** If true, this machine's queue is currently being executed. */
        this.lock_queue = false;
        /** List of current ticks per state. */
        this.clock_ = {};
        /**
         * Target object for the transitions, useful when composing the states
         * instance.
         */
        this.target = null;
        this.log_level_ = 0;
        /** Currently executing transition (if any) */
        this.transition = null;
        /** List of handlers receiving log messeges */
        this.log_handlers = [];
        /**
         * If true, the default log handler will be disabled - no `console.log`
         * messages.
         */
        this.def_log_handler_off = false;
        /**
         * Queue execution got postponed, because this machine is currently during
         * a transition from another machine's queue.
         */
        this.postponed_queue = false;
        this.internal_fields = [
            'states_all',
            'lock_queue',
            'states_active',
            'queue_',
            'lock',
            'last_promise',
            'log_level_',
            'clock_',
            'target',
            'internal_fields',
            'piped',
            'id_',
            'print_exception',
            'transition',
            'log_handlers',
            'postponed_queue',
            'def_log_handler_off'
        ];
        this.id_ = uuid();
        this.setTarget(target || this);
        if (register_all)
            this.registerAll();
        else
            this.register('Exception');
    }
    /**
     * All exceptions are caught into this state, including both synchronous and
     * asynchronous ones, coming from async methods and callbacks. You can
     * override it with your own and handle exceptions based on passed state
     * data and the actual state.
     *
     * @param err The exception object.
     * @param target_states Target states of the transition during
     * 	 which the exception was thrown.
     * @param base_states Base states in which the transition originated.
     * @param exception_src_handler The handler state which thrown the exception.
     * @param async_target_states Only for delayed mutations like
     *   [[addByCallback]], these are states which we're supposed to be activated
     *   by the callback.
     *
     * Example of a custom exception handler
     * ```
     * states = machine(['A', 'B', 'C'])
     * states.Exception_state = function(err, target_states) {
     *   // Re-adds the state 'C' in case of an exception when A is active
     *   if (exception_states.some((state) => state == 'C') && this.is('A')) {
     *     states.add 'C'
     *   }
     * }
     * ```
     *
     * Example of a manually thrown exception
     * ```
     * states.A_state = function(states) {
     *   foo = new SomeAsyncTask()
     *   foo.start()
     *   foo.once('error', (error) => {
     *     this.add('Exception', error, states)
     *   }
     * }
     * ```
     */
    Exception_state(err, target_states, base_states, exception_src_handler, async_target_states) {
        if (this.print_exception) {
            console.error('EXCEPTION from AsyncMachine');
        }
        if (target_states && target_states.length > 0) {
            this.log(`Exception \"${err}\" when setting states:\n` +
                `${target_states.join(', ')}`);
        }
        if (Array.isArray(base_states)) {
            this.log(`Source states of the transition were:\n` +
                `${base_states.join(', ') || '--EMPTY--'}`);
        }
        if (async_target_states && async_target_states.length > 0) {
            this.log(`Next states that were supposed to be (add|drop|set):\n` +
                `${async_target_states.join(', ')}`);
        }
        if (exception_src_handler) {
            console.error(`The call which caused the exception was ` + exception_src_handler);
        }
        // if the exception param was passed, print and throw (but outside of the
        // current stack trace)
        if (err) {
            if (this.print_exception) {
                console.error(err);
            }
            this.setImmediate(() => {
                throw err;
            });
        }
    }
    /**
     * Sets the target for the transition handlers. Useful to keep all your
     * methods in in one class while the states class is composed as an
     * attribute of the target object. There's also a shorthand for this method
     * as a [[AsyncMachine.constructor]]'s param.
     *
     * @param target Target object.
     *
     * ```
     * class Foo {
     *   constructor() {
     *     this.states = machine(['A', 'B', 'C'])
     *     this.states.setTarget(this)
     *     this.states.add('A')
     *   }
     *
     *   A_state() {
     *     console.log('State A set')
     *   }
     * }
     * ```
     */
    setTarget(target) {
        this.target = target;
        return this;
    }
    /**
     * Registers all defined states. Use it only if you don't define any other
     * attributes on the object (or it's prototype). If you do, register the
     * states manually with the [[register]] method. There's also a shorthand
     * for this method as [[AsyncMachine.constructor]]'s param.
     *
     * ```
     * class States extends AsyncMachine {
     *   constructor() {
     *     this.A = {}
     *     this.B = {}
     *
     *     this.registerAll()
     *     console.log(this.states_all) // -> ['Exception', 'A', 'B']
     *   }
     * }
     * ```
     */
    registerAll() {
        // test the instance vars
        for (let name in this) {
            let value = this[name];
            if (this.hasOwnProperty(name) &&
                !this.internal_fields.includes(name) &&
                !(value instanceof Function)) {
                this.register(name);
            }
        }
        // test the prototype chain
        var constructor = this.constructor.prototype;
        if (constructor === AsyncMachine.prototype)
            return;
        while (true) {
            for (let name in constructor) {
                let value = constructor[name];
                if (constructor.hasOwnProperty(name) &&
                    !this.internal_fields.includes(name) &&
                    !(value instanceof Function)) {
                    this.register(name);
                }
            }
            constructor = Object.getPrototypeOf(constructor);
            if (constructor === AsyncMachine.prototype)
                break;
        }
        // TODO validate relations
    }
    /**
     * Returns defined relations between two registered states.
     *
     * @param from_state
     * @param to_state
     * @returns List of relations.
     */
    getRelationsOf(from_state, to_state) {
        this.parseStates(from_state);
        if (to_state)
            this.parseStates(to_state);
        let state = this.get(from_state);
        let relations = [
            types_1.StateRelations.AFTER,
            types_1.StateRelations.ADD,
            types_1.StateRelations.DROP,
            types_1.StateRelations.REQUIRE
        ];
        return relations.filter(relation => {
            if (!state[relation])
                return false;
            // @ts-ignore
            if (to_state && !state[relation].includes(to_state)) {
                return false;
            }
            return true;
        });
    }
    is(states, tick) {
        if (!states) {
            return this.states_active;
        }
        let states_parsed = this.parseStates(states);
        const active = states_parsed.every(state => {
            return Boolean(this.states_active.includes(state));
        });
        if (!active) {
            return false;
        }
        if (states_parsed.length && tick !== undefined) {
            return this.clock(states) === tick;
        }
        return true;
    }
    /**
     * Returns `true` in case of all of the passed states aren't active.
     *
     * Example:
     *
     * ```
     * const example = machine(['A', 'B', 'C', 'D'])
     * example.add(['A', 'B'])
     *
     * // not(A) and not(C)
     * example.not(['A', 'C']) // -> false
     * // not(C) and not(D)
     * example.not(['C', 'D']) // -> true
     * ```
     *
     * @param states
     * @return
     */
    not(states) {
        let states_parsed = this.parseStates(states);
        return states_parsed.every(state => {
            return Boolean(!this.states_active.includes(state));
        });
    }
    any(...states) {
        return states.some(name => {
            return this.is(name);
        });
    }
    /**
     * Returns the current queue.
     *
     * To understand the returned format, refer to [[IQueueRow]].
     */
    queue() {
        return this.queue_;
    }
    /**
     * Simple method to check if a particular mutation has been queued. Returns
     * an index of the match or -1 in not found.
     *
     * All 3 params are checked and only the first match is returned, thus
     * this is not a full flaged solution for quering the queue.
     *
     * Example - check if `this.add('Bar')` is currently queued:
     * ```typescipt
     * import { machine, MutationTypes } from 'asyncmachine'
     * const example = machine(['Foo', 'Bar'])
     * example.Foo_state = function() {
     *   this.add('Bar') // -> null
     *   this.isQueued(MutationTypes.ADD, ['Bar']) // -> true
     * }
     * example.add('Foo') // -> true
     * ```
     *
     * @param mutation_type Type: add, drop or set
     * @param states List of states used in the mutation
     * @param target_machine Target to mutate. Defaults to `this.
     * @param without_params_only Matches only mutation without params.
     * @param states_strict_equal States of the mutation have to be exactly like
     *   `states`. If `false`, sets wider than (containing all of) the passed
     *   `states` will also be accepted.
     * @param start_index Start index for the lookup (in the queue)
     * @return Found index of -1
     *
     * TODO write a test
     */
    isQueued(mutation_type, states, target_machine = this, without_params_only = true, states_strict_equal = true, start_index = 0) {
        return this.queue().findIndex((r, index) => index >= start_index &&
            r[types_1.QueueRowFields.STATE_CHANGE_TYPE] == mutation_type &&
            r[types_1.QueueRowFields.TARGET] == target_machine &&
            ((without_params_only && !r[types_1.QueueRowFields.PARAMS].length) ||
                !without_params_only) &&
            // target states have to be at least as long as the checked ones
            // or exactly the same in case of a strict_equal
            ((states_strict_equal &&
                [types_1.QueueRowFields.STATES].length == states.length) ||
                (!states_strict_equal &&
                    [types_1.QueueRowFields.STATES].length >= states.length)) &&
            // and all the checked ones have to be included in the target ones
            states.every(s => r[types_1.QueueRowFields.STATES].includes(s)));
    }
    /**
     * Register passed state names. State properties should be already defined.
     *
     * @param states State names.
     * @return
     *
     * ```
     * const example = machine()
     * example.Enabled = {}
     * example.Disposed = { drop: ['Enabled'] }
     *
     * states.register('Enabled', 'Disposed')
     *
     * states.add('Enabled')
     * states.is() // -> 'Enabled'
     * ```
     */
    register(...states) {
        // TODO dont register during a transition
        for (let state of this.parseStates(states)) {
            // @ts-ignore Cant use this.get() here
            const exists = this[state];
            if (!exists) {
                console.error(`Missing state '${state}' in machine '${this.id()}'`);
            }
            if (!this.states_all.includes(state)) {
                this.states_all.push(state);
            }
            this.clock_[state] = 0;
        }
    }
    /**
     * TODO desc
     * TODO sample
     * TODO test
     * @param name
     */
    unregister(name) {
        throw new Error(`Not implemented - unregister('${name}')`);
        // TODO dont unregister during a transition
    }
    /**
     * Returns state's properties and relations.
     *
     * @param state State name.
     *
     * ```
     * states = machine(['A', 'B', 'C'])
     * states.A = { drop: ['B'] }
     *
     * states.get('A') // -> { drop: ['B'] }
     * ```
     */
    get(state) {
        if (!this.states_all.includes(state)) {
            throw new types_1.NonExistingStateError(this.id(true) + ':' + state);
        }
        return this[state];
    }
    set(target, states, ...params) {
        assert(target, 'First param required');
        if (!(target instanceof AsyncMachine)) {
            // TODO test for `states === 0`
            if (states !== undefined) {
                params = params ? [states, ...params] : [states];
            }
            states = target;
            target = this;
        }
        this.enqueue_(types_1.MutationTypes.SET, states, params, target);
        return this.processQueue_();
    }
    setByCallback(target, states, ...params) {
        assert(target, 'First param required');
        // TODO closure instead of bind
        return this.createCallback(this.createDeferred(this.set.bind(this), target, states, params));
    }
    setByListener(target, states, ...params) {
        assert(target, 'First param required');
        // TODO closure instead of bind
        return this.createListener(this.createDeferred(this.set.bind(this), target, states, params));
    }
    setNext(target, states, ...params) {
        assert(target, 'First param required');
        // TODO closure
        let fn = this.set.bind(this);
        return this.setImmediate(fn, target, states, params);
    }
    add(target, states, ...params) {
        assert(target, 'First param required');
        if (!(target instanceof AsyncMachine)) {
            // TODO test for `states === 0`
            if (states !== undefined) {
                params = params ? [states, ...params] : [states];
            }
            states = target;
            target = this;
        }
        this.enqueue_(types_1.MutationTypes.ADD, states, params, target);
        return this.processQueue_();
    }
    addByCallback(target, states, ...params) {
        assert(target, 'First param required');
        // TODO closure instead of bind
        return this.createCallback(this.createDeferred(this.add.bind(this), target, states, params));
    }
    addByListener(target, states, ...params) {
        assert(target, 'First param required');
        // TODO closure instead of bind
        return this.createListener(this.createDeferred(this.add.bind(this), target, states, params));
    }
    addNext(target, states, ...params) {
        assert(target, 'First param required');
        // TODO closure
        let fn = this.add.bind(this);
        return this.setImmediate(fn, target, states, params);
    }
    drop(target, states, ...params) {
        assert(target, 'First param required');
        if (!(target instanceof AsyncMachine)) {
            if (states !== undefined) {
                params = params ? [states, ...params] : [states];
            }
            states = target;
            target = this;
        }
        this.enqueue_(types_1.MutationTypes.DROP, states, params, target);
        return this.processQueue_();
    }
    dropByCallback(target, states, ...params) {
        assert(target, 'First param required');
        // TODO closure instead of bind
        return this.createCallback(this.createDeferred(this.drop.bind(this), target, states, params));
    }
    dropByListener(target, states, ...params) {
        assert(target, 'First param required');
        // TODO closure instead of bind
        return this.createListener(this.createDeferred(this.drop.bind(this), target, states, params));
    }
    dropNext(target, states, ...params) {
        assert(target, 'First param required');
        // TODO closure
        let fn = this.drop.bind(this);
        return this.setImmediate(fn, target, states, params);
    }
    /**
     * Pipes (forwards) a state to another machine.
     *
     * @param state Name of the state to pipe.
     * @param machine Target machine to which the state should be forwarded to.
     * @param target_state If the target state name should be different, this is
     *   that name. Applicable if only one state is being piped.
     * @param flags Different modes of piping. See [[PipeFlags]].
     *
     * Piping without negotiation
     * ```
     * const m1 = machine(['A', 'B', 'C'])
     * const m2 = machine(['A', 'B', 'C'])
     * m1.pipe('A', m2)
     * m1.add('A')
     * m2.is('A') // -> true
     * ```
     *
     * Piping with negotiation
     * ```
     * import { PipeFlags } from 'asyncmachine'
     * const m1 = machine(['A', 'B', 'C'])
     * const m2 = machine(['A', 'B', 'C'])
     * m2.A_enter = function() {
     *   return false
     * }
     * m1.pipe('A', m2, null, PipeFlags.NEGOTIATION)
     * m1.add('A')
     * m2.is('A') // -> false
     * ```
     */
    pipe(state, machine, target_state, flags = types_1.PipeFlags.FINAL) {
        this.pipeBind(state, machine, target_state, flags);
    }
    /**
     * Pipes all the states from this machine to the `target`. Doesn't pipe the
     * `Exception` state.
     *
     * The exception state is never piped.
     *
     * @param target Target machine to which the state should be forwarded.
     */
    pipeAll(target, flags = types_1.PipeFlags.FINAL) {
        // Do not forward the Exception state
        let states_all = this.states_all.filter(state => state !== 'Exception');
        this.pipeBind(states_all, target, null, flags);
    }
    /**
     * Removes an existing pipe. All params are optional.
     *
     * @param states Source states. Empty means any state.
     * @param machine Target machine. Empty means any machine.
     * @param flags Pipe flags. Empty means any flags.
     *
     * TODO optimise, if needed
     */
    pipeRemove(states, machine, flags = types_1.PipeFlags.FINAL) {
        assert(!!flags, `Flags can't be 0`);
        // Add defaults when a sole INVERT requested
        if (flags == types_1.PipeFlags.INVERT) {
            flags = types_1.PipeFlags.INVERT | types_1.PipeFlags.FINAL;
        }
        let bindings = flags ? this.getPipeBindings(flags) : null;
        let event_types = flags && bindings ? Object.keys(bindings) : null;
        let parsed_states = states ? this.parseStates(states) : null;
        let to_emit = [];
        for (let [state, pipes] of Object.entries(this.piped)) {
            pipes = pipes;
            // TODO remove casting once Object.entries() is typed correctly
            if (parsed_states &&
                !parsed_states.includes(state))
                continue;
            for (let i = 0; i < pipes.length; i++) {
                let pipe = pipes[i];
                if (machine && machine !== pipe.machine)
                    continue;
                if (event_types && !event_types.includes(pipe.event_type))
                    continue;
                this.removeListener(`${state}_${pipe.event_type}`, pipe.listener);
                pipes.splice(i, 1);
                // stay on the same index
                i--;
                if (!to_emit.includes(pipe.machine)) {
                    to_emit.push(pipe.machine);
                }
                if (!to_emit.includes(this)) {
                    to_emit.push(this);
                }
            }
            if (!pipes.length) {
                delete this.piped[state];
            }
        }
        for (let machine of to_emit) {
            // TODO emit pipe-in-removed pipe-out-removed, passing the pipe binding
            machine.emit('pipe');
        }
    }
    /**
     * TODO should remove a binding returned by pipe() and pipeAll() methods
     */
    pipeRemoveBinding( /*binding*/) {
        throw new Error('TODO');
    }
    /**
     * Returns the current tick of the passed state.
     *
     * State's clock starts with 0 and on each activation gets incremented
     * by 1. Ticks lets you keep control flow's integrity across async listeners,
     * by aborting them once the state changes. Easiest way to get the tick-based
     * abort function is to use [[getAbort]].
     *
     * @param state Name of the state
     * @return Current tick of the passed state
     *
     * Example
     * ```
     * const example = machine(['A', 'B', 'C'])
     * example.add('A')
     * example.add('A')
     * example.clock('A') // -> 1
     * example.drop('A')
     * example.add('A')
     * example.clock('A') // -> 2
     * ```
     */
    clock(state) {
        // @ts-ignore
        return this.clock_[state] || 0;
    }
    /**
     * Creates a prototype child with it's own active states, clock, queue and
     * locks.
     *
     * Useful for creating new instances of machines created using the
     * [[machine]] factory in an efficient manner.
     *
     * Example
     * ```
     * const parent = machine(['A', 'B', 'C'])
     * const child = parent.createChild()
     *
     * child.add('A')
     * child.is() // -> ['A']
     * parent.is() // -> []
     * ```
     *
     * // TODO write a test
     */
    createChild() {
        // TODO check if during a transition
        const child = Object.create(this);
        child.states_active = [];
        child.states_all = [...this.states_all];
        child.clock_ = {};
        child.queue_ = [];
        child.lock = false;
        child.lock_queue = false;
        child.postponed_queue = false;
        child._events = null;
        child.piped = {};
        child.transition = null;
        for (let state of this.states_all) {
            child.clock[state] = 0;
        }
        return child;
    }
    /**
     * Indicates if this instance is currently during a transition.
     *
     * When a machine is during a transition, all state changes will be queued
     * and executed once the transition and the previously queued state
     * changes are finished. See [[queue]].
     *
     * Example
     * ```
     * const example = machine(['A', 'B', 'C'])
     *
     * example.A_enter = function() {
     *   this.duringTransition() // -> true
     * }
     *
     * example.A_state = function() {
     *   this.duringTransition() // -> true
     * }
     *
     * example.add('A')
     * ```
     */
    duringTransition() {
        return this.lock;
    }
    /**
     * Returns the list of active states from which the current transition
     * started.
     *
     * Requires [[duringTranstion]] to be true or it'll throw.
     */
    from() {
        if (!this.transition || this.transition.machine !== this) {
            throw new Error(`[AsyncMachine] ${this.id()} not during an (own) transition`);
        }
        // TODO write a test
        return this.transition.before;
    }
    /**
     * Returns the list of target states which are about to be active after the
     * transition finishes.
     *
     * Requires [[duringTranstion]] to be true or it'll throw.
     */
    to() {
        if (!this.transition || this.transition.machine !== this) {
            throw new Error(`[AsyncMachine] ${this.id()} not during an (own) transition`);
        }
        // TODO write a test
        return this.transition.states;
    }
    /**
     * Returns an abort function, based on the current [[clock]] tick of the
     * passed state. Optionally allows to next an existing abort function.
     *
     * The abort function return a boolean `true` in case the flow for the
     * specific state should be aborted, because:
     * - the state has been de-activated (at least once)
     * - the nested abort function returns `true`
     *
     * Example
     * ```
     * const example = machine(['A', 'B', 'C'])
     *
     * example.A_state = function() {
     *   const abort = this.getAbort('A')
     *   setTimeout( () => {
     *     if (abort()) return
     *     console.log('never reached')
     *   }, 0)
     * }
     *
     * example.add('A')
     * example.drop('A')
     * ```
     *
     * TODO support multiple states
     *
     * @param state Name of the state
     * @param abort Existing abort function (optional)
     * @return A new abort function
     */
    getAbort(state, abort) {
        var tick = this.clock(state);
        return this.getAbortFunction(state, tick, abort);
    }
    /**
     * Resolves the returned promise when all of the passed states are active
     * (at the same time). Accepts an optional abort function.
     *
     * Example
     * ```
     * const example = machine(['A', 'B', 'C'])
     * example.when(['A', 'B']).then( () => {
     *   console.log('A, B')
     * }
     *
     * example.add('A')
     * example.add('B') // prints 'A, B'
     * ```
     *
     * TODO support push cancellation
     *
     * @param states List of state names
     * @param abort Existing abort function (optional)
     * @return Promise resolved once all states are set simultaneously.
     */
    when(states, abort) {
        let states_parsed = this.parseStates(states);
        return new Promise(resolve => {
            // early resolve if all the states are currently set
            if (this.is(states_parsed)) {
                return resolve();
            }
            this.bindToStates(states_parsed, resolve, abort);
        });
    }
    /**
     * Resolves the returned promise when all of the passed states are NOT active
     * (at the same time). Accepts an optional abort function.
     *
     * Example
     * ```
     * const example = machine(['A', 'B', 'C'])
     * example.when(['A', 'B']).then( () => {
     *   console.log('A, B')
     * }
     *
     * example.add('A')
     * example.add('B') // prints 'A, B'
     * ```
     *
     * TODO support push cancellation
     * TODO merge with `when(active, inactive)`
     *
     * @param states List of state names
     * @param abort Existing abort function (optional)
     * @return Promise resolved once all states are set simultaneously.
     */
    whenNot(states, abort) {
        let states_parsed = this.parseStates(states);
        return new Promise(resolve => {
            // early resolve if all the states are NOT currently set
            if (this.not(states_parsed)) {
                return resolve();
            }
            this.bindToNotStates(states_parsed, resolve, abort);
        });
    }
    logLevel(log_level) {
        if (log_level !== undefined) {
            this.print_exception = Boolean(log_level);
            this.log_level_ = parseInt(log_level, 10);
            return this;
        }
        else {
            return this.log_level_;
        }
    }
    id(id) {
        if (id === true) {
            return (this.id_ || '')
                .replace(/[^\w\d]/g, '-')
                .replace(/-+/g, '-')
                .toLocaleLowerCase();
        }
        if (id !== undefined) {
            if (id != this.id_) {
                let old_id = this.id_;
                this.id_ = id;
                this.emit('id-changed', id, old_id);
            }
            return this;
        }
        else
            return this.id_;
    }
    // @ts-ignore
    on(event, listener, context) {
        // if event is a NAME_state event, fire immediately if the state is set
        if ((event.slice(-6) === '_state' || event.slice(-6) === '_enter') &&
            this.is(event.slice(0, -6))) {
            this.catchPromise(listener.call(context));
            // if event is a NAME_end event, fire immediately if the state isnt set
        }
        else if ((event.slice(-4) === '_end' &&
            !this.is(event.slice(0, -4))) ||
            (event.slice(-5) === '_exit' &&
                !this.is(event.slice(0, -5)))) {
            this.catchPromise(listener.call(context));
        }
        super.on(event, listener, context);
        return this;
    }
    // @ts-ignore
    once(event, listener, context) {
        // is event is a NAME_state event, fire immediately if the state is set
        // and dont register the listener
        if ((event.slice(-6) === '_state' || event.slice(-6) === '_enter') &&
            this.is(event.slice(0, -6))) {
            this.catchPromise(listener.call(context));
            // is event is a NAME_end event, fire immediately if the state is not set
            // and dont register the listener
        }
        else if ((event.slice(-4) === '_end' &&
            !this.is(event.slice(0, -4))) ||
            (event.slice(-5) === '_exit' &&
                !this.is(event.slice(0, -5)))) {
            this.catchPromise(listener.call(context));
        }
        else {
            super.once(event, listener, context);
        }
        return this;
    }
    /**
     * Binds the Exception state to the promise error handler. Handy when working
     * with promises.
     *
     * See [[Exception_state]].
     *
     * @param promise The promise to handle
     * @param target_states States for which the promise was created (the
     *   one that failed).
     * @return The source promise, for piping.
     */
    catchPromise(promise, target_states) {
        if (isPromise(promise)) {
            promise.catch((error) => {
                this.add('Exception', error, target_states);
            });
        }
        return promise;
    }
    /**
     * Diffs two state sets and returns the ones present only in the first one.
     *
     * @param states1 List of source states.
     * @param states2 List of states to diff against (picking up the non
     *   existing ones).
     * @return List of states in states1, but not in states2.
     */
    diffStates(states1, states2) {
        return states1.filter(name => !states2.includes(name));
    }
    /**
     * Push a new message to the log with an optional level (defaults to 1).
     *
     * @param {string} msg
     * @param {number} level
     */
    log(msg, level = 1) {
        for (const handler of this.log_handlers) {
            handler(msg, level);
        }
        if (level > this.log_level_ || this.def_log_handler_off)
            return;
        let prefix = this.id() ? `[${this.id()}] ` : '';
        msg = prefix + msg;
        console.log(msg);
    }
    /**
     * Returns a string representation of the machine using the [[statesToString]]
     * method.
     */
    toString() {
        return this.statesToString();
    }
    /**
     * Returns a string representation of the machine (name and the list of
     * states), optionally including the in-active states.
     *
     * @param include_inactive
     */
    statesToString(include_inactive = false) {
        let ret = `${this.id()}`;
        let footer = '';
        for (const state of this.states_all) {
            if (this.states_active.includes(state)) {
                ret += `\n    ${state} (${this.clock(state)})`;
            }
            else if (include_inactive) {
                footer += `\n    -${state} (${this.clock(state)})`;
            }
        }
        return ret + footer;
    }
    // PRIVATES
    getPipeBindings(flags = types_1.PipeFlags.FINAL) {
        const pf = types_1.PipeFlags;
        flags = flags || pf.FINAL;
        const bindings = {};
        if (flags & pf.INVERT || flags & pf.FINAL_ENTER) {
            // enter
            if (flags & pf.NEGOTIATION || flags & pf.NEGOTIATION_ENTER) {
                bindings['enter'] = 'drop';
            }
            if (flags & pf.FINAL || flags & pf.FINAL_ENTER) {
                bindings['state'] = 'drop';
            }
            // exit
            if (flags & pf.FINAL || flags & pf.FINAL_EXIT) {
                bindings['end'] = 'add';
            }
            if (flags & pf.NEGOTIATION || flags & pf.NEGOTIATION_EXIT) {
                bindings['exit'] = 'add';
            }
        }
        else {
            // enter
            if (flags & pf.NEGOTIATION || flags & pf.NEGOTIATION_ENTER) {
                bindings['enter'] = 'add';
            }
            if (flags & pf.FINAL || flags & pf.FINAL_ENTER) {
                bindings['state'] = 'add';
            }
            // exit
            if (flags & pf.FINAL || flags & pf.FINAL_EXIT) {
                bindings['end'] = 'drop';
            }
            if (flags & pf.NEGOTIATION || flags & pf.NEGOTIATION_EXIT) {
                bindings['exit'] = 'drop';
            }
        }
        // @ts-ignore
        return bindings;
    }
    getPipeFlagsLabels(flags) {
        const names = Object.keys(types_1.PipeFlags).filter(key => !isNaN(Number(types_1.PipeFlags[key])));
        const ret = [];
        for (const flag of names) {
            if (flags & types_1.PipeFlags[flag]) {
                ret.push(types_1.PipeFlagsLabels[flag]);
            }
        }
        return ret;
    }
    // TODO overload signatures for the optional requested_state
    // TODO rename machine to target and target to machine, as its confusing
    pipeBind(states, machine, requested_state, flags = types_1.PipeFlags.FINAL) {
        const pf = types_1.PipeFlags;
        assert(!!flags, `Flags can't be 0`);
        // Add defaults when a sole INVERT requested
        if (flags == pf.INVERT) {
            flags = pf.INVERT | pf.FINAL;
        }
        let bindings = this.getPipeBindings(flags);
        let parsed_states = this.parseStates(states);
        if (requested_state && typeof requested_state !== 'string')
            throw new Error('target_state has to be string or null');
        const is_negotiation = bindings['enter'] || bindings['exit'];
        if (is_negotiation && flags & pf.LOCAL_QUEUE) {
            throw new Error('Cant pipe with negotiation into a local queue');
        }
        let tags = this.getPipeFlagsLabels(flags).join(':');
        if (parsed_states.length == 1 && parsed_states[0] == requested_state)
            this.log(`[pipe:${tags}] '${parsed_states[0]}' as '${requested_state}' to '${machine.id()}'`, 2);
        else
            this.log(`[pipe:${tags}] '${parsed_states.join(', ')}' to '${machine.id()}'`, 2);
        let emit_on = [];
        for (let state of parsed_states) {
            // accept a different name only when one state is piped
            let target_state = (parsed_states.length == 1 && requested_state) || state;
            for (let [event_type, method_name] of Object.entries(bindings)) {
                // TODO extract
                let pipe_listener = () => {
                    let target = flags && flags & types_1.PipeFlags.LOCAL_QUEUE ? this : machine;
                    if (this.transition) {
                        this.transition.addStep([machine.id(true), target_state], [this.id(true), state], types_1.TransitionStepTypes.PIPE);
                    }
                    const ret = target[method_name](machine, target_state);
                    // return only for negotiation pipes
                    if (['enter', 'exit'].includes(event_type)) {
                        // TODO check both lock of both of the machines
                        if (target.duringTransition()) {
                            this.log(`[pipe] Negotiation not possible, machine ${target.id(true)} during a transition`, 2);
                        }
                        // TODO add a canceled step in case of negotiation and a negative
                        // result
                        if (ret === false) {
                            this.log(`[pipe:cancelled] State '${state}' not accepted by the ` +
                                `machine '${target.id()}'`);
                        }
                        return ret;
                    }
                };
                // TODO extract
                // TODO check for duplicates
                if (!this.piped[state]) {
                    this.piped[state] = [];
                }
                // @ts-ignore
                this.piped[state].push({
                    state: target_state,
                    machine: machine,
                    event_type: event_type,
                    flags,
                    listener: pipe_listener
                });
                // assert target states
                machine.parseStates(target_state);
                // setup the forwarding listener
                // TODO listener-less piping
                // read from #pipes directly, inside the transition
                this.on(`${state}_${event_type}`, pipe_listener);
            }
            if (!emit_on.includes(this)) {
                emit_on.push(this);
            }
            if (machine !== this && !emit_on.includes(machine)) {
                emit_on.push(machine);
            }
        }
        for (let machine of emit_on) {
            // TODO emit pipe-in and pipe-out, on source and target respectively
            // passing the pipe binding
            machine.emit('pipe');
        }
    }
    /**
     * Override for EventEmitter method calling a specific listener. Binds to
     * a promise if returned by the listener.
     *
     * TODO incorporate into EE, saving one call stack frame
     */
    callListener(listener, context, params) {
        var ret = listener.apply(context, params);
        // assume params[0] are the target states of the transition
        return this.catchPromise(ret, params[0]);
    }
    // TODO make it cancellable
    setImmediate(fn, ...params) {
        if (setImmediate) {
            return setImmediate.apply(null, [fn].concat(params));
        }
        else {
            return setTimeout(fn.apply(null, params), 0);
        }
    }
    /**
     * Return `true` if the current state differs from the `states_before`.
     *
     * @param states_before List of previously active states.
     */
    hasStateChanged(states_before) {
        const length_equals = this.is().length === states_before.length;
        return (!length_equals ||
            Boolean(this.diffStates(states_before, this.is()).length));
    }
    /**
     * Parse state names, check if they exist and always return an array. Throws
     * in of an error.
     *
     * @param states State names to check
     * @return An array of valid state names.
     */
    parseStates(states) {
        // TODO remove duplicates
        const states_parsed = [].concat(states);
        return states_parsed.filter(state => {
            // @ts-ignore Cant use this.get() here
            const exists = this[state];
            if (typeof state !== 'string' || !exists) {
                let id = this.id() ? ` in the machine "${this.id()}"` : '';
                throw new types_1.NonExistingStateError(`State "${state}" doesn't exist${id}`);
            }
            return true;
        });
    }
    /**
     * Returns the JSON structure of states along with their relations.
     *
     * @return JSON version of the current machine. Valid input for the
     *   [[machine]] factory.
     */
    states() {
        let ret = {};
        for (let state of this.states_all) {
            ret[state] = this.get(state);
        }
        return ret;
    }
    /*
       * Puts a transition in the queue, handles a log msg and unifies the states
       * array.
       *
       * TODO generic for the machine param?
       */
    enqueue_(type, states, params = [], target = this) {
        const type_label = types_1.MutationTypes[type].toLowerCase();
        const states_parsed = target.parseStates(states);
        // Detect duplicates and avoid queueing them.
        if (!params.length &&
            this.detectQueueDuplicates_(target, type, states_parsed)) {
            this.log(`[queue:skipped] Duplicate detected for [${types_1.MutationTypes[type].toLowerCase()}] '${states_parsed.join(', ')}'`, 2);
            return;
        }
        let queue = this.queue_;
        if (this.duringTransition()) {
            if (this.transition && this.transition.source_machine !== this) {
                // TODO log msg for using the parent queue
                queue = this.transition.source_machine.queue_;
            }
            if (target !== this) {
                this.log(`[queue:${type_label}] [${target.id()}]` +
                    ` ${states_parsed.join(', ')}`, 2);
            }
            else {
                this.log(`[queue:${type_label}] ${states_parsed.join(', ')}`, 2);
            }
        }
        queue.push([type, states_parsed, params, false, target]);
    }
    /**
     * Detect queue duplicates for mutations without params.
     *
     * 1. Check if a mutation is scheduled
     * 2. Check if a counter mutation isn't scheduled later
     *
     * @param machine The machine which queue is being checked
     * @param target Target to mutate
     * @param mutation_type
     * @param states
     * @private
     *
     * TODO write a test
     */
    detectQueueDuplicates_(target, mutation_type, states) {
        // check if this mutation is already scheduled
        const index = this.isQueued(mutation_type, states, target);
        if (index == -1) {
            return false;
        }
        let counter_mutation_type;
        switch (mutation_type) {
            case types_1.MutationTypes.ADD:
                counter_mutation_type = types_1.MutationTypes.DROP;
                break;
            case types_1.MutationTypes.DROP:
                counter_mutation_type = types_1.MutationTypes.ADD;
                break;
            default:
            case types_1.MutationTypes.SET:
                // avoid duplicating `set` only if at the end of the queue
                return index && this.queue().length - 1;
        }
        // Check if a counter mutation is scheduled and broaden the match
        // - with or without params
        // - state sets same or bigger than `states`
        return this.isQueued(counter_mutation_type, states, target, false, false, index + 1);
    }
    // Goes through the whole queue collecting return values.
    processQueue_() {
        if (!this.queue_.length)
            return false;
        let queued = false;
        if (this.lock_queue) {
            this.log(`[postpone] postponing the transition, the queue is running`, 3);
            queued = true;
        }
        else if (this.lock) {
            // instance is during a transition from an external queue
            // wait for it to finish OR schedule this queue somehow
            // 			this.log(`[abort] Aborting queue processing, machine locked`, 3)
            this.log(`[postpone] postponing the transition, machine locked by another` +
                ` one`, 3);
            queued = true;
            if (!this.postponed_queue) {
                this.postponed_queue = true;
                this.once('queue-end', () => {
                    this.postponed_queue = false;
                    this.log('[resume] resuming an aborted queue', 3);
                    this.processQueue_();
                });
            }
        }
        if (queued) {
            this.emit('queue-changed');
            // queued mutation requests return `null`
            // TODO write a test
            return null;
        }
        let ret = [];
        this.lock_queue = true;
        let row;
        while ((row = this.queue_.shift())) {
            if (!row[types_1.QueueRowFields.TARGET]) {
                row[types_1.QueueRowFields.TARGET] = this;
            }
            this.transition = new transition_1.default(this, row);
            // TODO do the queue checks here, per each entry and postpone if busy
            // expose the current transition also on the target machine
            row[types_1.QueueRowFields.TARGET].transition = this.transition;
            ret.push(this.transition.exec());
            // GC the transition
            row[types_1.QueueRowFields.TARGET].transition = null;
        }
        // GC the transition
        this.transition = null;
        this.lock_queue = false;
        this.emit('queue-end');
        return ret[0] || false;
    }
    createDeferred(fn, target, states, state_params) {
        // TODO use the current transition's states if available (for enter/exit
        // transitions)
        var transition_states = this.is();
        var params = [target];
        if (states) {
            params.push(states);
        }
        if (state_params.length) {
            params.push.apply(params, state_params);
        }
        const async_states = [].concat(params[0] instanceof AsyncMachine ? params[1] : params[0]);
        const gc = function () {
            // GC
            deferred = null;
            resolve = null;
            reject = null;
        };
        let deferred = new types_1.Deferred();
        let resolve = deferred.resolve;
        let reject = deferred.reject;
        deferred.resolve = callback_params => {
            try {
                const ret = fn.apply(null, params.concat(callback_params));
                if (resolve)
                    resolve(ret);
            }
            catch (err) {
                const ret = this.add('Exception', err, transition_states, async_states);
                if (reject)
                    reject(ret);
            }
            finally {
                gc();
            }
        };
        deferred.reject = err => {
            const ret = this.add('Exception', err, transition_states, async_states);
            if (reject)
                reject(ret);
            gc();
        };
        this.last_promise = deferred.promise;
        return deferred;
    }
    createCallback(deferred) {
        return (err = null, ...params) => {
            if (err) {
                return deferred.reject(err);
            }
            else {
                return deferred.resolve(params);
            }
        };
    }
    createListener(deferred) {
        return (...params) => deferred.resolve(params);
    }
    /**
     * Sets the new active states incrementing the counters.
     *
     * @return An array of previously active states.
     */
    setActiveStates_(explicite_states, target) {
        let previous = this.states_active;
        let new_states = this.diffStates(target, this.states_active);
        let removed_states = this.diffStates(this.states_active, target);
        let nochange_states = this.diffStates(target, new_states);
        this.states_active = target;
        // Tick all the new states and the explicit multi states
        for (let state of target) {
            let data = this.get(state);
            if (!previous.includes(state) ||
                (explicite_states.includes(state) && data.multi)) {
                // @ts-ignore
                this.clock_[state]++;
            }
        }
        // construct a logging msg
        const log_msg = [];
        if (new_states.length) {
            log_msg.push('+' + new_states.join(' +'));
        }
        if (removed_states.length) {
            log_msg.push('-' + removed_states.join(' -'));
        }
        // TODO fix
        if (nochange_states.length && this.log_level_ > 2) {
            if (new_states.length || removed_states.length) {
                log_msg.push('\n    ');
            }
            log_msg.push(nochange_states.join(', '));
        }
        if (log_msg.length) {
            this.log('[state] ' + log_msg.join(' '), 1);
        }
        return previous;
    }
    /**
     * Returns an object for a given handler.
     * @param name Handler's name
     */
    getMethodContext(name) {
        if (this.target &&
            this.target[name] &&
            this.target[name] instanceof Function) {
            return this.target;
        }
        else if (this[name] && this[name] instanceof Function) {
            return this;
        }
        return null;
    }
    // TODO bind to _enter and _exit as well to support the negotiation phase in
    // piped events
    bindToStates(states, listener, abort) {
        const enter = () => {
            let should_abort = abort && abort();
            if (!should_abort && this.is(states))
                listener();
            if (this.is(states)) {
                this.log(`[bind:off] ${states.join(', ')}`, 3);
                for (let state of states)
                    this.removeListener(`${state}_state`, enter);
            }
        };
        this.log(`[bind:on] ${states.join(', ')}`, 3);
        for (let state of states)
            this.on(`${state}_state`, enter);
    }
    // TODO bind to _enter and _exit as well to support the negotiation phase in
    // piped events
    // TODO merge with #with as `.with(['A', 'B'], ['C']) runs when `+A+B-C`
    bindToNotStates(states, listener, abort) {
        const exit = () => {
            let should_abort = abort && abort();
            if (!should_abort && this.not(states))
                listener();
            if (this.not(states)) {
                this.log(`[bind:off] ${states.join(', ')}`, 3);
                for (let state of states)
                    this.removeListener(`${state}_end`, exit);
            }
        };
        this.log(`[bind:on] ${states.join(', ')}`, 3);
        for (let state of states)
            this.on(`${state}_end`, exit);
    }
    // TODO compose abort functions without recursion
    getAbortFunction(state, tick, abort) {
        return () => {
            let quit = false;
            let msg = '';
            if (abort && abort()) {
                quit = true;
                msg = 'the parent abort function has quit';
            }
            else if (!this.is(state)) {
                quit = true;
                msg = 'the state is not set';
            }
            else if (!this.is(state, tick)) {
                quit = true;
                msg = 'the has tick changed';
            }
            if (quit) {
                const full_msg = `[abort] ${state}; reason - ${msg}\n\t${this.is().join(', ')}`;
                this.log(full_msg, 1);
                if (this.log_level_ > 2 || this.log_handlers.length) {
                    this.log(new Error().stack);
                    console.trace(full_msg);
                }
                return true;
            }
            return false;
        };
    }
}
exports.default = AsyncMachine;
exports.AsyncMachine = AsyncMachine;
function isPromise(promise) {
    return promise && promise.then && promise.catch;
}
//# sourceMappingURL=asyncmachine.js.map