import AsyncMachine from './asyncmachine';
import Transition from './transition';
export declare type BaseStates = 'Exception';
export declare type TAsyncMachine = AsyncMachine<any, IBind, IEmit>;
export interface IBind {
    (event: 'tick', listener: (before: string[]) => boolean | void, context?: Object): this;
    (event: 'id-changed', listener: (new_id: string, old_id: string) => boolean | void, context?: Object): this;
    (event: 'transition-init', listener: (transition: Transition) => boolean | void, context?: Object): this;
    (event: 'transition-start', listener: (transition: Transition) => boolean | void, context?: Object): this;
    (event: 'transition-end', listener: (transition: Transition) => boolean | void, context?: Object): this;
    (event: 'transition-step', listener: (...steps: ITransitionStep[]) => boolean | void, context?: Object): this;
    (event: 'pipe', listener: (pipe: TPipeBindings) => boolean | void, context?: Object): this;
    (event: 'state-registered', listener: (state: string) => boolean | void, context?: Object): this;
    (event: 'state-deregistered', listener: (state: string) => boolean | void, context?: Object): this;
    (event: 'transition-cancelled', listener: (transition: Transition) => boolean | void, context?: Object): this;
    (event: 'queue-changed', listener: () => boolean | void, context?: Object): this;
    (event: 'queue-end', listener: () => boolean | void, context?: Object): this;
    (event: 'Exception_enter', listener: (err: Error, target_states: string[], base_states: string[], exception_transition: string, async_target_states?: string[]) => boolean | void, context?: Object): this;
    (event: 'Exception_state', listener: (err: Error, target_states: string[], base_states: string[], exception_transition: string, async_target_states?: string[]) => any, context?: Object): this;
    (event: 'Exception_exit', listener: () => boolean | void, context?: Object): this;
    (event: 'Exception_end', listener: () => any, context?: Object): this;
    (event: 'Exception_Any', listener: () => boolean | void, context?: Object): this;
    (event: 'Any_Exception', listener: () => boolean | void, context?: Object): this;
    (event: 'ts-dynamic', listener: Function): this;
}
export interface IEmit {
    (event: 'tick', before: string[]): boolean;
    (event: 'id-changed', new_id: string, old_id: string): boolean;
    (event: 'transition-init', transition: Transition): boolean;
    (event: 'transition-start', transition: Transition): boolean;
    (event: 'transition-end', transition: Transition): boolean;
    (event: 'transition-step', ...steps: ITransitionStep[]): boolean;
    (event: 'pipe'): boolean;
    (event: 'state-registered', state: string): boolean;
    (event: 'state-deregistered', state: string): boolean;
    (event: 'transition-cancelled', transition: Transition): boolean;
    (event: 'queue-changed'): boolean;
    (event: 'Exception_enter', err: Error, target_states: string[], base_states: string[], exception_transition: string, async_target_states?: string[]): boolean;
    (event: 'Exception_state', err: Error, target_states: string[], base_states: string[], exception_transition: string, async_target_states?: string[]): any;
    (event: 'Exception_exit'): boolean;
    (event: 'Exception_end'): boolean;
    (event: 'Exception_Any'): boolean;
    (event: 'Any_Exception'): boolean;
    (event: 'queue-end'): boolean;
    (event: 'ts-dynamic', ...params: any[]): boolean;
}
export interface IState<T extends string> {
    /** Decides about the order of activations (transitions) */
    after?: (T | BaseStates)[];
    /** When set, sets also the following states */
    add?: (T | BaseStates)[];
    /** When set, blocks activation (or deactivates) given states */
    drop?: (T | BaseStates)[];
    /** State will be rejected if any of those aren't set */
    require?: (T | BaseStates)[];
    /** When true, the state will be set automatically, if it's not blocked */
    auto?: boolean;
    /**
     * Multi states always triggers the enter and state transitions, plus
     * the clock is always incremented
     */
    multi?: boolean;
}
export declare enum MutationTypes {
    DROP = 0,
    ADD = 1,
    SET = 2,
}
/**
 * Queue enum defining param positions in queue's entries.
 */
export declare enum QueueRowFields {
    STATE_CHANGE_TYPE = 0,
    STATES = 1,
    PARAMS = 2,
    AUTO = 3,
    TARGET = 4,
}
export interface IQueueRow {
    0: MutationTypes;
    1: string[];
    2: any[];
    3: boolean;
    4: TAsyncMachine;
}
export declare class Deferred {
    promise: Promise<any>;
    resolve: (...params: any[]) => void;
    reject: (err?: any) => void;
    constructor();
}
export declare enum StateRelations {
    AFTER = "after",
    ADD = "add",
    REQUIRE = "require",
    DROP = "drop",
}
export declare enum TransitionStepTypes {
    RELATION = 1,
    TRANSITION = 4,
    SET = 8,
    DROP = 16,
    NO_SET = 32,
    REQUESTED = 64,
    CANCEL = 128,
    PIPE = 256,
}
export declare enum StateStructFields {
    MACHINE_ID = 0,
    STATE_NAME = 1,
}
export interface IStateStruct {
    0: string;
    1: string;
}
export declare enum TransitionStepFields {
    STATE = 0,
    SOURCE_STATE = 1,
    TYPE = 2,
    DATA = 3,
}
export interface ITransitionStep {
    0: IStateStruct;
    1?: IStateStruct;
    2?: TransitionStepTypes;
    3?: any;
}
export declare type TAbortFunction = () => boolean;
export declare type TStateAction = 'add' | 'drop' | 'set';
export declare type TStateMethod = 'enter' | 'exit' | 'state' | 'end';
export interface IPipeNegotiationBindings {
    enter: TStateAction;
    exit: TStateAction;
}
export interface IPipeStateBindings {
    state: TStateAction;
    end: TStateAction;
}
export declare type TPipeBindings = IPipeStateBindings | IPipeNegotiationBindings;
export interface IPipedStateTarget {
    state: string;
    machine: TAsyncMachine;
    event_type: TStateMethod;
    listener: Function;
    flags?: PipeFlags;
}
/**
 * By default piped are "_state" and "_end" methods, not the negotiation ones.
 * Use the PipeFlags.NEGOTIATION flag to pipe "_enter" and "_exit" methods, and
 * thus, to participate in the state negotiation. This mode DOES NOT guarantee,
 * that the state was successfully negotiated in the source machine.
 *
 * To invert the state, use the PipeFlags.INVERT flag.
 *
 * To append the transition to the local queue (instead of the target
 * machine's one), use the PipeFlags.LOCAL_QUEUE. This will alter the
 * transition order.
 */
export declare enum PipeFlags {
    NEGOTIATION = 1,
    INVERT = 4,
    LOCAL_QUEUE = 8,
    FINAL = 16,
    NEGOTIATION_ENTER = 32,
    NEGOTIATION_EXIT = 64,
    FINAL_ENTER = 128,
    FINAL_EXIT = 256,
}
export declare const PipeFlagsLabels: {
    NEGOTIATION: string;
    INVERT: string;
    LOCAL_QUEUE: string;
    FINAL: string;
    NEGOTIATION_ENTER: string;
    NEGOTIATION_EXIT: string;
    FINAL_ENTER: string;
    FINAL_EXIT: string;
};
export declare class TransitionException extends Error {
    err: Error;
    transition: string;
    constructor(err: Error, transition: string);
}
export declare class NonExistingStateError extends Error {
    constructor(name: string);
}
export declare type TLogHandler = (msg: string, level?: number) => any;
